from typing import Any

from fastapi import APIRouter

from ...context import context
from ...models import Artist
from ...models.response_types import ArtistInfo
from ..servers.artists import Server

app = APIRouter(prefix="/artists")
server = Server()


@app.get("/")
async def load_saved_artists() -> list[dict[str, Any]]:
    artists = server.load_saved_artists()
    return list(artists)


@app.get("/search")
async def search_artist(name: str) -> list[dict[str, Any]]:
    artists = server.create_artist_search_results(name)
    return list(artists)


@app.get("/add")
async def save_new_artist(id: str, name: str) -> None:
    artist = Artist(id, name)
    return context.storage.save_new_artist(artist)


@app.get("/toggle")
async def toggle_artist_type(id: str) -> None:
    artists: list[Artist] = context.storage.artists
    for artist in artists:
        if artist.id == id:
            artist.toggle_type()
    context.storage.artists = artists


@app.get("/recommendations")
async def recommend_artists() -> list[ArtistInfo]:
    return server.create_artist_recommendations()



from functools import cached_property

from package_utils.context import Context as Context_

from ..clients import spotify
from ..models import Config, Options, Secrets
from ..storage.storage import Storage


class Context(Context_[Options, Config, Secrets]):
    @cached_property
    def spotify_client(self) -> spotify.Client:
        return spotify.Client(self.secrets)

    @cached_property
    def storage(self) -> Storage:
        return Storage()


context = Context(Options, Config, Secrets)



import io
import os
import urllib.parse
import zipfile
from dataclasses import dataclass, field

import cli
import requests
from plib import Path

from music.webapp.context import context

from .frontend_release import FrontendRelease


def load_dist_path() -> Path:
    domain_name = urllib.parse.urlparse(context.config.hostname).netloc
    return Path("/") / "var" / "www" / domain_name


@dataclass
class Frontend:
    content_path: Path = field(default_factory=load_dist_path)

    @classmethod
    def check_content(cls) -> None:
        frontend = Frontend()
        if context.options.update_frontend or not frontend.content_path.exists():
            frontend.update_content()

    def update_content(self) -> None:
        frontend_release = FrontendRelease.load()
        if self.content_path.mtime < frontend_release.mtime:
            self.download_content(frontend_release)

    def download_content(self, frontend_release: FrontendRelease) -> None:
        if not self.content_path.exists():
            self.create_content_path()
        content = requests.get(frontend_release.browser_download_url).content
        content_io = io.BytesIO(content)
        zip_file = zipfile.ZipFile(content_io)
        with zip_file:
            zip_file.extractall(self.content_path)
        self.content_path.mtime = frontend_release.mtime

    def create_content_path(self) -> None:
        try:
            username = os.getlogin()
        except OSError:  # pragma: nocover (on GitHub action)
            username = "runner"  # pragma: nocover
        cli.run("mkdir", self.content_path, root=True)
        cli.run(f"chown -R {username}:{username}", self.content_path, root=True)



import typing
from typing import TypeVar

import plib
from simple_classproperty import classproperty

T = TypeVar("T", bound="Path")


class Path(plib.Path):
    @classmethod
    @classproperty
    def source_root(cls: type[T]) -> T:
        return cls(__file__).parent.parent

    @classmethod
    @classproperty
    def assets(cls: type[T]) -> T:
        path = cls.script_assets / cls.source_root.parent.name / cls.source_root.name
        return typing.cast(T, path)

    @classmethod
    @classproperty
    def config(cls: type[T]) -> T:
        path = cls.assets / "config" / "config.yaml"
        return typing.cast(T, path)



from __future__ import annotations

from dataclasses import dataclass
from datetime import datetime
from typing import TypeVar

import dacite
import requests
from package_utils.dataclasses.mixins import SerializationMixin

from music.webapp.context import context

T = TypeVar("T")


@dataclass
class FrontendRelease(SerializationMixin):
    updated_at: str
    browser_download_url: str

    @classmethod
    def load(cls) -> FrontendRelease:
        release_url = f"https://api.github.com/repos/{context.config.frontend_repository}/releases/latest"
        frontend_info = requests.get(release_url).json()["assets"][0]
        config = dacite.Config(strict=False)
        return cls.from_dict(frontend_info, config=config)

    @property
    def mtime(self) -> float:
        format_str = "%Y-%m-%dT%H:%M:%SZ"
        return datetime.strptime(self.updated_at, format_str).timestamp()



import functools
from collections.abc import Callable
from typing import Any, TypeVar

T = TypeVar("T")


def combine_chunks(f: Callable[..., list[T]]) -> Callable[..., list[T]]:
    """
    Combine function calls on chunks instead of whole array in one time :return: All
    results of the chunks.
    """

    @functools.wraps(f)
    def chunked_wrapper(*args: Any, **kwargs: Any) -> list[T]:
        self, items, *args_tuple = args

        @combine_offsets
        def calculate_chunk(self: Any, offset: int, limit: int) -> list[T]:
            return f(self, items[offset : offset + limit], *args_tuple, **kwargs)

        return calculate_chunk(self, len(items))

    return chunked_wrapper


def combine_offsets(
    f: Callable[..., list[T]], chunk_size: int = 50
) -> Callable[..., list[T]]:
    """
    Combine function calls with offset and limit instead of total amount :param
    chunk_size: Size of the chunks :return: All results of the chunks.
    """

    def chunked_wrapper(self: Any, amount: int, *args: Any, **kwargs: Any) -> list[T]:
        return [
            item
            for offset in range(0, amount, chunk_size)
            for item in f(self, *args, offset=offset, limit=chunk_size, **kwargs)
        ]

    return chunked_wrapper



import cli

from ..context import context
from .artist import ArtistUpdater


def check_for_new_songs() -> None:
    description = "Checking for new songs"
    cli.console.clear_live()
    artists = cli.progress(  # type: ignore
        context.storage.artists, description=description, unit="artists", cleanup=True
    )
    for artist in artists:
        ArtistUpdater(artist).check_for_new_songs()



from music.context import context

from .. import updaters
from ..models import Path


def main() -> None:
    """
    Download new songs.
    """
    if context.options.clean_download_ids:
        updaters.download_ids.clean_download_ids()
    else:
        if context.options.add is not None:
            tracks = [context.options.add]
            updaters.tracks.add_tracks_by_name(tracks)
        collect_new_songs()


def collect_new_songs() -> None:
    upload_to_phone = should_upload_to_phone()
    if not upload_to_phone and not context.storage.tracks_to_download:
        updaters.artists.check_for_new_songs()
    if context.storage.tracks_to_download:
        download_new_songs()
        upload_to_phone = should_upload_to_phone()
    if upload_to_phone:
        upload_new_downloads()


def should_upload_to_phone() -> bool:
    processed_songs_present = not Path.processed_songs.is_empty()
    return context.options.upload_to_phone and processed_songs_present


def download_new_songs() -> None:
    # lazy import for performance
    from ..download import download_new_songs

    download_new_songs.download_new_songs()


def upload_new_downloads() -> None:
    # lazy import for performance
    from . import uploader

    uploader.start()



from collections.abc import Iterator

import cli

from ..context import context
from ..models.response_types import Track


def add_tracks_by_name(names: list[str]) -> None:
    tracks_iterator = generate_new_tracks_by_name(names)
    tracks = list(tracks_iterator)
    if tracks:
        context.storage.save_new_tracks(tracks)


def generate_new_tracks_by_name(names: list[str]) -> Iterator[Track]:
    for name in names:
        tracks = context.spotify_client.search_song(name)
        yield from process_search_results(tracks)


def process_search_results(tracks: list[Track]) -> Iterator[Track]:
    show_results = True
    while show_results and tracks:
        track = tracks.pop(0)
        message = f"{track.full_name}\nDownload?"
        if cli.confirm(message):  # type: ignore
            yield track
            show_results = False
        else:
            show_results = cli.confirm("See next result?")  # type: ignore



from dataclasses import dataclass

from .path import Path


@dataclass
class Config:
    secrets_path = Path.secrets
    download_chunk_size: int = 20
    download_retries: int = 5
    phone_upload_port: int = 2222
    frontend_port: int = 8080



from dataclasses import dataclass

from .path import Path


@dataclass
class Options:
    config_path: Path = Path.config
    headless: bool = False
    debug: bool = False
    restart: bool = False
    update_frontend: bool = False
    backend: bool = False



from dataclasses import dataclass


@dataclass
class Config:
    app_name: str = "music.server:app"
    hostname: str = "https://music.com"
    frontend_repository: str = "quintenroets/music-frontend"
    backend_port: int = 13000
    session_name: str | None = None



import calendar
from dataclasses import dataclass
from datetime import datetime
from functools import cached_property

from mutagen import oggopus

from ..models import Path


def run() -> None:
    for path in Path.downloaded_songs.glob("*.opus"):
        if path.size == 0:
            raise Exception(f"{path} is empty file")
        DownloadedTrackProcessor(path).run()


@dataclass
class DownloadedTrackProcessor:
    path: Path
    set_title: bool = True

    def __post_init__(self) -> None:
        self.metadata = oggopus.OggOpus(self.path)  # type: ignore

    def run(self) -> None:
        if self.set_title:
            self.add_title_metadata()
        # do this after all other operations to avoid resetting mtime
        self.path.mtime = max(self.mtime.timestamp(), 0)
        self.move_to_processed_songs()

    def move_to_processed_songs(self) -> None:
        self.path.rename(Path.processed_songs / self.path.name)

    def add_title_metadata(self) -> None:
        month = calendar.month_name[self.mtime.month][:3]
        date_ = f"{month} {self.mtime.day}, {self.mtime.year}"
        title = self.metadata["title"][0]
        self.metadata["title"] = f"{title} | {date_}"
        self.metadata.save()

    @cached_property
    def mtime(self) -> datetime:
        date = self.metadata["date"][0]
        parts = date.split("-")
        if len(parts) == 1:
            parts += [1, 1]  # pragma: nocover

        y, m, d = parts
        y, m, d = int(y), int(m), int(d)
        return datetime(y, m, d)



from collections.abc import Iterator

import cli

from music.webapp.context import context

from .frontend import Frontend


def main() -> None:
    start_backend()
    open_frontend()


def open_frontend() -> None:
    Frontend.check_content()
    if not context.options.headless:
        cli.urlopen(context.config.hostname)


def start_backend() -> None:
    commands = generate_commands()
    cli.run_commands(*commands)


def generate_commands() -> Iterator[str]:
    backend_command = "music-backend"
    session_name = context.config.session_name or backend_command
    if context.options.restart:
        yield f"tmux kill-session -t {session_name}"

    options = "--debug" if context.options.debug else "--no-debug"
    yield f"tmux new-session -s {session_name} -d {backend_command} {options}"



from ..context import context
from ..utils.batched import batched
from . import spotdl
from .downloader import Downloader


def download_new_songs() -> None:
    fix_song_count()
    size = context.config.download_chunk_size
    ids_to_download = context.storage.ids_to_download
    for ids in batched(ids_to_download, size=size):
        Downloader(ids).start()


def fix_song_count() -> None:
    num_songs = len(context.storage.tracks_to_download)
    progress_handler = spotdl.downloader.progress_handler
    progress_handler.set_song_count(num_songs)
    progress_handler.set_song_count = lambda count: None



from collections.abc import Awaitable, Callable

import tbhandler
from fastapi import FastAPI, Request, Response
from fastapi.middleware.cors import CORSMiddleware
from fastapi.responses import PlainTextResponse

from ..context import context
from .routers import artists, songs


def create_app() -> FastAPI:
    app = FastAPI()

    origins = [f"http://localhost:{context.config.frontend_port}"]
    app.add_middleware(
        CORSMiddleware,
        allow_origins=origins,
        allow_credentials=True,
        allow_headers=["*"],
    )

    @app.middleware("http")
    async def exception_handling(
        request: Request, call_next: Callable[[Request], Awaitable[Response]]
    ) -> Response:
        try:
            response = await call_next(request)
        except Exception as exception:
            tbhandler.show(exit_after=False, repeat=False)
            response = PlainTextResponse(str(exception), status_code=500)
        return response

    routers = artists.app, songs.app
    for router in routers:
        app.include_router(router)
    return app



from collections.abc import Iterator
from functools import cached_property

import cli
from package_utils.storage import cached_path_property

from ..models import Artist, Path
from ..models.response_types import Track


class Storage:
    downloaded_tracks: dict[str, str] = Path.download_ids.cached_content
    tracks_to_download: dict[str, str] = Path.to_download.cached_content
    recommendation_frequencies: dict[str, float] = Path.recommendations.cached_content

    @property
    def artists(self) -> list[Artist]:
        return self._artists

    @artists.setter
    def artists(self, value: list[Artist]) -> None:
        self._artists = value  # type: ignore

    @property
    @cached_path_property(Path.artists)
    def _artists(self) -> list[Artist]:
        return [Artist.from_dict(artist) for artist in Path.artists.yaml]

    @_artists.fget.setter  # noqa
    def _artists(self, artists: list[Artist]) -> None:
        # use sort_index explicitly because dataclass ordering does not work
        artists = sorted(artists, key=lambda artist: artist.sort_index)
        Path.artists.yaml = [artist.dict() for artist in artists]

    def save_new_artist(self, artist: Artist) -> None:
        self.artists = self.artists + [artist]

    @property
    @cached_path_property(Path.artists)
    def artist_ids(self) -> set[str]:
        return set(artist.id for artist in self.artists)

    @property
    @cached_path_property(Path.artists)
    def artists_per_id(self) -> dict[str, Artist]:
        return {artist.id: artist for artist in self.artists}

    def get_artist(self, id: str) -> Artist:
        return self.artists_per_id[id]

    @cached_property
    def downloaded_track_names(self) -> set[str]:
        names = self.downloaded_tracks.values()
        return set(names)

    @cached_property
    def downloaded_track_ids(self) -> set[str]:
        ids = self.downloaded_tracks.keys()
        return set(ids)

    @property
    def ids_to_download(self) -> list[str]:
        ids = self.tracks_to_download.keys()
        return list(ids)

    def save_new_tracks(self, tracks: list[Track]) -> None:
        tracks_to_download = self._extract_tracks_to_download(tracks)
        tracks_mapping = {track.id: track.full_name for track in tracks_to_download}
        if tracks_mapping:
            self.tracks_to_download |= tracks_mapping

    def _extract_tracks_to_download(self, tracks: list[Track]) -> Iterator[Track]:
        tracks = sorted(tracks, key=lambda track_: track_.popularity, reverse=True)
        for track in tracks:
            if self._should_download(track):
                cli.console.print(track.full_name)
                yield track

    def _should_download(self, track: Track) -> bool:
        """
        A new track has a non-existing:
        - id
        - full name
        because the same song can occur with:
        - different id's: different album
        - different full name: different artists
        """
        return (
            track.should_download
            and track.id not in self.downloaded_track_ids
            and track.full_name not in self.downloaded_track_names
        )



import typing

from ..models import Artist, Path


class Paths:
    def __init__(self, artist: Artist) -> None:
        self.name = artist.name.replace("/", "")

    @property
    def albums(self) -> Path:
        return self.create_path("albums")

    @property
    def albums_count(self) -> Path:
        return self.create_path("albums_count")

    @property
    def album_counts(self) -> Path:
        return self.create_path("album_counts")

    @property
    def top_tracks(self) -> Path:
        return self.create_path("top_tracks")

    def create_path(self, name: str) -> Path:
        path = Path.cache_assets / name / self.name
        path = path.with_suffix(".yaml")
        return typing.cast(Path, path)



from collections.abc import Iterator

from ..context import context


def clean_download_ids() -> None:
    track_info_iterator = generate_tracks_with_unique_name()
    ids, names = zip(*track_info_iterator)
    tracks = context.spotify_client.songs(ids)
    downloaded_tracks = {track.id: name for track, name in zip(tracks, names)}
    context.storage.downloaded_tracks = downloaded_tracks


def generate_tracks_with_unique_name() -> Iterator[tuple[str, str]]:
    encountered_names = set()
    for id_, name in context.storage.downloaded_tracks.items():
        if name not in encountered_names:
            encountered_names.add(name)
            yield id_, name



import uvicorn

from music.webapp.context import context

from ...models import Path


def main() -> None:
    reload_dirs = str(Path.source_root) if context.options.debug else None
    log_level = "info" if context.options.debug else "warning"
    uvicorn.run(
        context.config.app_name,
        port=context.config.backend_port,
        reload_dirs=reload_dirs,
        reload=context.options.debug,
        log_level=log_level,
    )



import random
from collections.abc import Iterator
from dataclasses import dataclass
from itertools import islice
from typing import Any

from ...context import context
from ...models.response_types import ArtistInfo


@dataclass
class Server:
    number_of_recommendations: int = 50
    max_new_recommendation_tries: int = 10

    @classmethod
    def create_artist_search_results(cls, name: str) -> Iterator[dict[str, Any]]:
        saved_artist_ids = context.storage.artist_ids
        artists = context.spotify_client.search_artist(name)
        for artist in artists:
            is_added = artist.id in saved_artist_ids
            yield artist.dict() | {"added": is_added}

    def create_artist_recommendations(self) -> list[ArtistInfo]:
        recommendations = self._create_artist_recommendations()
        self.increase_recommendation_frequencies(recommendations)
        frequencies = context.storage.recommendation_frequencies
        return sorted(recommendations, key=lambda artist: frequencies[artist.id])

    def _create_artist_recommendations(self) -> set[ArtistInfo]:
        iterator = self.generate_recommended_artists()
        limited_iterator = islice(iterator, self.number_of_recommendations)
        return set(limited_iterator)

    def generate_recommended_artists(self) -> Iterator[ArtistInfo]:
        saved_artist_ids = list(context.storage.artist_ids)
        random.shuffle(saved_artist_ids)
        seed_ids = saved_artist_ids[: self.max_new_recommendation_tries]
        for id_ in seed_ids:
            related_artists = context.spotify_client.related_artists(id_)
            for artist in related_artists:
                if artist.id not in saved_artist_ids:
                    yield artist

    @classmethod
    def increase_recommendation_frequencies(cls, artists: set[ArtistInfo]) -> None:
        frequencies = context.storage.recommendation_frequencies
        for artist in artists:
            frequencies[artist.id] = frequencies.get(artist.id, 0) + 1
        context.storage.recommendation_frequencies = frequencies

    @classmethod
    def load_saved_artists(cls) -> Iterator[dict[str, str]]:
        artists = context.storage.artists
        artist_ids = list(context.storage.artist_ids)
        api_infos = context.spotify_client.artists(artist_ids)
        for artist, info in zip(artists, api_infos):
            yield info.dict() | artist.dict()



import os
from collections.abc import Iterator
from dataclasses import dataclass
from functools import cached_property

import cli
import hostfinder
import pysftp

from ..context import context
from ..download.downloaded_songs_processor import DownloadedTrackProcessor
from ..models import Path


def start() -> None:  # pragma: nocover
    with cli.status("Looking for phone"):  # type: ignore
        ip = hostfinder.find_host(port=context.config.phone_upload_port)
    if ip is not None:
        Uploader(ip).run()


@dataclass
class Uploader:  # pragma: nocover
    ip: str

    def __post_init__(self) -> None:
        with cli.status("Connecting to phone"):  # type: ignore
            self.phone_connection = self.create_phone_connection()

    @property
    def phone_connection_options(self) -> pysftp.CnOpts:
        connection_options = pysftp.CnOpts()
        connection_options.hostkeys = None
        connection_options.log = True
        return connection_options

    def create_phone_connection(self) -> pysftp.Connection:
        username = None if "GITHUB_ACTIONS" in os.environ else os.getlogin()
        return pysftp.Connection(
            self.ip,
            port=context.config.phone_upload_port,
            username=username,
            password=context.secrets.phone_connection,
            cnopts=self.phone_connection_options,
        )

    def run(self) -> None:
        with self.phone_connection:
            self._run()

    def _run(self) -> None:
        self.phone_connection.makedirs(Path.phone)
        if context.options.fix_mtimes_on_phone:
            self.fix_mtimes_on_phone()
        self.process_remote_deletes()
        self.upload_new_songs()

    def fix_mtimes_on_phone(self) -> None:
        for path in Path.all_songs.iterdir():
            DownloadedTrackProcessor(path, set_title=False).run()

        description = "Fixing mtime of all songs on phone"
        song_paths = Path.all_songs.iterdir()
        song_paths = self.iterate_over_song_paths(song_paths, description=description)

        for path in song_paths:
            self.fix_mtime_on_phone(path)

    def fix_mtime_on_phone(self, path: Path) -> None:
        remote_path = f"{Path.phone}/{path.name}"
        mtime = self.phone_connection.stat(remote_path).st_mtime
        if mtime is None or int(mtime) != int(path.mtime):
            fixed_mtime = max(path.mtime, 0)
            fixed_mtimes = (fixed_mtime, fixed_mtime)
            self.phone_connection._sftp.utime(remote_path, fixed_mtimes)  # type: ignore

    @classmethod
    def iterate_over_all_songs(cls, description: str) -> Iterator[Path]:
        paths = Path.all_songs.iterdir()
        paths_with_length = list(paths)
        yield from cli.progress(  # type: ignore
            paths_with_length, description=description, unit="songs", cleanup=True
        )

    def process_remote_deletes(self) -> None:
        songs = Path.all_songs.iterdir()
        for song in songs:
            if song.name not in self.uploaded_song_names:
                print(f"Removing {song.stem}")
                song.rename(Path.deleted / song.name)

    @cached_property
    def uploaded_song_names(self) -> set[str]:
        with cli.console.status("Reading songs on phone"):
            uploaded_names = self.phone_connection.listdir(Path.phone)
        return set(uploaded_names)

    def upload_new_songs(self) -> None:
        description = "Copying new songs to phone"
        new_song_paths = Path.processed_songs.glob("*.opus")
        new_song_paths = self.iterate_over_song_paths(
            new_song_paths, description=description
        )
        for path in new_song_paths:
            self.upload_song(path)

    @classmethod
    def iterate_over_song_paths(
        cls, paths: Iterator[Path], description: str
    ) -> Iterator[Path]:
        paths_with_length = list(paths)
        yield from cli.progress(  # type: ignore
            paths_with_length, description=description, unit="songs", cleanup=True
        )

    def upload_song(self, path: Path) -> None:
        self.phone_connection.put(
            localpath=str(path),
            remotepath=f"{Path.phone}/{path.name}",
            preserve_mtime=True,
        )
        path.rename(Path.all_songs / path.name)



import logging
import os
import random
import time
from dataclasses import dataclass

import requests
import spotipy
from retry import retry
from spotipy.cache_handler import CacheFileHandler
from spotipy.oauth2 import SpotifyClientCredentials

from ..models import Path, Secrets


@dataclass
class Spotify(spotipy.Spotify):  # type: ignore
    secrets: Secrets
    market: str = "BE"

    def __post_init__(self) -> None:
        ccm = SpotifyClientCredentials(
            client_id=self.secrets.spotify.client_id,
            client_secret=self.secrets.spotify.client_secret,
            cache_handler=CacheFileHandler(Path.cache),
        )
        super().__init__(
            client_credentials_manager=ccm,
            requests_timeout=1,
            status_retries=0,
            retries=3,
            backoff_factor=1,
        )
        logging.disable()  # ignore retry error log messages from spotify library

    @retry(requests.exceptions.ReadTimeout, tries=10)
    def _internal_call(
        self, method: str, url: str, payload: dict[str, str], params: dict[str, str]
    ) -> dict[str, str] | None:
        try:
            for market_param in ("country", "market"):
                if market_param in params:
                    params[market_param] = self.market
            response = super()._internal_call(method, url, payload, params)
        except spotipy.exceptions.SpotifyException as e:
            if e.http_status in (404, 429):
                # include 404 status because spotify api sometimes returns 404 error
                # when it should be 429
                # https://community.spotify.com/t5/Spotify-for-Developers/
                # Intermittent-404s-Getting-Playlist-Tracks-via-API/m-p/5356770
                self.sleep()  # pragma: nocover
                raise requests.exceptions.ReadTimeout  # pragma: nocover
            else:
                raise
        return response  # type: ignore

    @classmethod
    def sleep(cls) -> None:  # pragma: nocover
        time.sleep(2)
        if "GITHUB_ACTION" in os.environ:
            # Unit tests lead to rate limits because all requests start at the same time
            extra_time = 5 + random.Random().random() * 10
            time.sleep(extra_time)



from dataclasses import dataclass
from enum import Enum
from typing import Any

import dacite
from package_utils.dataclasses.mixins import SerializationMixin, T


class ArtistType(Enum):
    NORMAL = "normal"
    FAVORITE = "favorite"


@dataclass(order=True)
class Artist(SerializationMixin):
    id: str
    name: str
    type_: ArtistType = ArtistType.NORMAL

    @classmethod
    def from_dict(
        cls: type[T], items: dict[str, Any], config: dacite.Config | None = None
    ) -> T:
        config = dacite.Config(type_hooks={ArtistType: ArtistType})
        if "type" in items:
            items["type_"] = items.pop("type")
        return super().from_dict(items, config)

    def dict(self) -> dict[str, str]:
        items = super().dict()
        items["type"] = items.pop("type_").value
        return items

    @property
    def sort_index(self) -> tuple[bool, str]:
        """
        Favorites first and then order by name.
        """
        return self.is_normal, self.name

    def toggle_type(self) -> None:
        self.type_ = ArtistType.FAVORITE if self.is_normal else ArtistType.NORMAL

    @property
    def is_normal(self) -> bool:
        return self.type_ == ArtistType.NORMAL
